@@ -17,7 +17,6 @@
 import datetime
 import os
 import re
-import string
 from typing import List, Tuple, Optional
 
 bullets = '+-*'
@@ -30,6 +29,8 @@ under2_regex = re.compile(r'^-+\s*$')
 bullet_regex = re.compile(r'^[-+*]')
 linkid_regex = re.compile(r'^\[(?P<link_id>\S*)]:\s*(?P<link>.*)')
 
+default_header = '# Changelog\n\nAll notable changes to this project will be documented in this file'
+
 
 def _strip_link(token):
     if link_literal := re.fullmatch(r'\[(.*?)]\((.*?)\)', token):
@@ -47,7 +48,7 @@ def _join_markdown(segments: List[str]) -> str:
     text: List[str] = []
     last_bullet = False
     for segment in segments:
-        is_bullet = bullet_regex.match(segment) and '\n' not in segment
+        is_bullet = bullet_regex.match(segment)
 
         if not is_bullet or not last_bullet:
             text.append('')
@@ -62,18 +63,38 @@ def _join_markdown(segments: List[str]) -> str:
 class VersionEntry:
     def __init__(self):
         self.sections = {'': []}
-        self.name: str = ''
+        self.name: str = 'Unreleased'
         self.date: Optional[datetime.date] = None
         self.tags: List[str] = []
         self.link: str = ''
-        self.link_id: str = None
+        self.link_id: str = ''
         self.line_no: int = -1
 
-    def __str__(self) -> str:
-        if self.link:
-            segments = [f'[{self.name}]']
+    def body(self, md: bool = True) -> str:
+        segments = []
+
+        for section, entries in self.sections.items():
+            if section:
+                if md:
+                    segments.append(f'### {section.title()}')
+                else:
+                    segments.append(f'{section.upper()}:')
+
+            if len(entries) > 0:
+                segments.append(_join_markdown(entries))
+
+        return _join_markdown(segments)
+
+    def header(self, md: bool = True) -> str:
+        segments = []
+
+        if md:
+            segments.append('##')
+
+        if self.link and md:
+            segments.append(f'[{self.name}]')
         else:
-            segments = [self.name]
+            segments.append(self.name)
 
         if self.date or len(self.tags) > 0:
             segments.append('-')
@@ -85,28 +106,36 @@ class VersionEntry:
 
         return ' '.join(segments)
 
+    def text(self, md: bool = True) -> str:
+        return self.header(md) + '\n\n' + self.body(md)
+
+    def __str__(self) -> str:
+        return self.header(False)
+
 
 class Changelog:
-    def __init__(self, path: os.PathLike):
-        self.path = path
-        self.header = ''
-        self.versions = []
+    def __init__(self, path: os.PathLike = None):
+        self.path: os.PathLike = path
+        self.header: str = ''
+        self.versions: List[VersionEntry] = []
+        self.links = {}
+
+        if not os.path.exists(path):
+            self.header = default_header
+            return
 
         # Read file
         with open(path, 'r') as fp:
-            self.lines = fp.readlines()
+            lines = fp.readlines()
 
         section = ''
         in_block = False
         in_code = False
 
-        self.links = {}
-
-        links = {}
         segments: List[Tuple[int, List[str], str]] = []
         header_segments = []
 
-        for line_no, line in enumerate(self.lines):
+        for line_no, line in enumerate(lines):
             if in_code:
                 # this is the contents of a code block
                 segments[-1][1].append(line)
@@ -143,7 +172,7 @@ class Changelog:
 
             elif match := linkid_regex.match(line):
                 # this is a link definition in the form '[id]: link', so add it to the link table
-                links[match['link_id'].lower()] = match['link']
+                self.links[match['link_id'].lower()] = match['link']
 
             elif line.isspace():
                 # skip empty lines
@@ -216,18 +245,17 @@ class Changelog:
             if match := re.fullmatch(r'\[(.*)]', version.name):
                 # ref-matched link
                 link_id = match[1].lower()
-                if link_id in links:
-                    version.link = links.pop(link_id)
+                if link_id in self.links:
+                    version.link = self.links.pop(link_id)
                     version.link_id = None
                     version.name = match[1]
 
-            elif version.link_id in links:
+            elif version.link_id in self.links:
                 # id-matched link
-                version.link = links.pop(version.link_id)
+                version.link = self.links.pop(version.link_id)
 
         # strip whitespace from header
         self.header = _join_markdown(header_segments)
-        self.links = links
 
     def write(self, path: os.PathLike = None):
         if path is None:
@@ -236,23 +264,18 @@ class Changelog:
         v_links = {}
         v_links.update(self.links)
 
+        segments = [self.header]
+
+        for version in self.versions:
+            if version.link:
+                v_links[version.name] = version.link
+
+            segments.append(version.text())
+
+        for link_id, link in v_links.items():
+            segments.append(f'[{link_id.lower()}]: {link}')
+
+        text = _join_markdown(segments)
+
         with open(path, 'w') as fp:
-            fp.write(self.header)
-            fp.write('\n\n')
-
-            for version in self.versions:
-                fp.write(f'## {version}\n\n')
-
-                if version.link:
-                    v_links[version.name] = version.link
-
-                for section in version.sections:
-                    if section:
-                        fp.write(f'### {section}\n\n')
-
-                    if len(version.sections[section]) > 0:
-                        fp.write(_join_markdown(version.sections[section]))
-                        fp.write('\n\n')
-
-            for link_id, link in v_links.items():
-                fp.write(f'[{link_id.lower()}]: {link}\n')
+            fp.write(text)
